@@ -7,8 +7,9 @@ use syn::{
 };
 
 use crate::{
-    diagnostic::{DiagnosticConcreteArgs, DiagnosticDef, DiagnosticDefArgs},
-    utils::forward_to_single_field_variant,
+    diagnostic::{DiagnosticConcreteArgs, DiagnosticDef},
+    forward::WhichFn,
+    utils::gen_all_variants_with,
 };
 
 #[derive(Debug)]
@@ -48,41 +49,24 @@ impl Parse for Code {
 
 impl Code {
     pub(crate) fn gen_enum(variants: &[DiagnosticDef]) -> Option<TokenStream> {
-        let code_pairs = variants.iter().map(
-            |DiagnosticDef {
-                 ident,
-                 fields,
-                 args,
-             }| {
-                match args {
-                    DiagnosticDefArgs::Transparent => {
-                        Some(forward_to_single_field_variant(ident, fields, quote! { code() }))
+        gen_all_variants_with(
+            variants,
+            WhichFn::Code,
+            |ident, fields, DiagnosticConcreteArgs { code, .. }| {
+                let code = &code.as_ref()?.0;
+                Some(match fields {
+                    syn::Fields::Named(_) => {
+                        quote! { Self::#ident { .. } => std::option::Option::Some(std::boxed::Box::new(#code)), }
                     }
-                    DiagnosticDefArgs::Concrete(DiagnosticConcreteArgs { code, .. }) => {
-                        let code = &code.as_ref()?.0;
-                        Some(match fields {
-                            syn::Fields::Named(_) => {
-                                quote! { Self::#ident { .. } => std::option::Option::Some(std::boxed::Box::new(#code)), }
-                            }
-                            syn::Fields::Unnamed(_) => {
-                                quote! { Self::#ident(..) => std::option::Option::Some(std::boxed::Box::new(#code)), }
-                            }
-                            syn::Fields::Unit => {
-                                quote! { Self::#ident => std::option::Option::Some(std::boxed::Box::new(#code)), }
-                            }
-                        })
+                    syn::Fields::Unnamed(_) => {
+                        quote! { Self::#ident(..) => std::option::Option::Some(std::boxed::Box::new(#code)), }
                     }
-                }
+                    syn::Fields::Unit => {
+                        quote! { Self::#ident => std::option::Option::Some(std::boxed::Box::new(#code)), }
+                    }
+                })
             },
-        );
-        Some(quote! {
-            fn code<'a>(&'a self) -> std::option::Option<std::boxed::Box<dyn std::fmt::Display + 'a>> {
-                match self {
-                    #(#code_pairs)*
-                    _ => std::option::Option::None,
-                }
-            }
-        })
+        )
     }
 
     pub(crate) fn gen_struct(&self) -> Option<TokenStream> {
