@@ -1,12 +1,15 @@
 use proc_macro2::TokenStream;
-use quote::{format_ident, quote};
+use quote::quote;
 use syn::{punctuated::Punctuated, DeriveInput, Token};
 
 use crate::code::Code;
 use crate::diagnostic_arg::DiagnosticArg;
+use crate::forward::{Forward, WhichFn};
 use crate::help::Help;
+use crate::label::Labels;
+use crate::related::Related;
 use crate::severity::Severity;
-use crate::snippets::Snippets;
+use crate::source_code::SourceCode;
 use crate::url::Url;
 
 pub enum Diagnostic {
@@ -30,8 +33,27 @@ pub struct DiagnosticDef {
 }
 
 pub enum DiagnosticDefArgs {
-    Transparent,
-    Concrete(DiagnosticConcreteArgs),
+    Transparent(Forward),
+    Concrete(Box<DiagnosticConcreteArgs>),
+}
+
+impl DiagnosticDefArgs {
+    pub(crate) fn forward_or_override_enum(
+        &self,
+        variant: &syn::Ident,
+        which_fn: WhichFn,
+        mut f: impl FnMut(&DiagnosticConcreteArgs) -> Option<TokenStream>,
+    ) -> Option<TokenStream> {
+        match self {
+            Self::Transparent(forward) => Some(forward.gen_enum_match_arm(variant, which_fn)),
+            Self::Concrete(concrete) => f(concrete).or_else(|| {
+                concrete
+                    .forward
+                    .as_ref()
+                    .map(|forward| forward.gen_enum_match_arm(variant, which_fn))
+            }),
+        }
+    }
 }
 
 #[derive(Default)]
@@ -39,8 +61,11 @@ pub struct DiagnosticConcreteArgs {
     pub code: Option<Code>,
     pub severity: Option<Severity>,
     pub help: Option<Help>,
-    pub snippets: Option<Snippets>,
+    pub labels: Option<Labels>,
+    pub source_code: Option<SourceCode>,
     pub url: Option<Url>,
+    pub forward: Option<Forward>,
+    pub related: Option<Related>,
 }
 
 impl DiagnosticConcreteArgs {
@@ -54,11 +79,15 @@ impl DiagnosticConcreteArgs {
         let mut severity = None;
         let mut help = None;
         let mut url = None;
+        let mut forward = None;
         for arg in args {
             match arg {
                 DiagnosticArg::Transparent => {
                     return Err(syn::Error::new_spanned(attr, "transparent not allowed"));
                 }
+                DiagnosticArg::Forward(to_field) => {
+                    forward = Some(to_field);
+                }
                 DiagnosticArg::Code(new_code) => {
                     // TODO: error on multiple?
                     code = Some(new_code);
@@ -74,13 +103,18 @@ impl DiagnosticConcreteArgs {
                 }
             }
         }
-        let snippets = Snippets::from_fields(fields)?;
+        let labels = Labels::from_fields(fields)?;
+        let source_code = SourceCode::from_fields(fields)?;
+        let related = Related::from_fields(fields)?;
         let concrete = DiagnosticConcreteArgs {
             code,
             help,
+            related,
             severity,
-            snippets,
+            labels,
             url,
+            forward,
+            source_code,
         };
         Ok(concrete)
     }
@@ -92,14 +126,15 @@ impl DiagnosticDefArgs {
         fields: &syn::Fields,
         attr: &syn::Attribute,
         allow_transparent: bool,
-    ) -> Result<Self, syn::Error> {
+    ) -> syn::Result<Self> {
         let args =
             attr.parse_args_with(Punctuated::<DiagnosticArg, Token![,]>::parse_terminated)?;
         if allow_transparent
             && args.len() == 1
             && matches!(args.first(), Some(DiagnosticArg::Transparent))
         {
-            return Ok(Self::Transparent);
+            let forward = Forward::for_transparent_field(fields)?;
+            return Ok(Self::Transparent(forward));
         } else if args.iter().any(|x| matches!(x, DiagnosticArg::Transparent)) {
             return Err(syn::Error::new_spanned(
                 attr,
@@ -114,7 +149,7 @@ impl DiagnosticDefArgs {
             .into_iter()
             .filter(|x| !matches!(x, DiagnosticArg::Transparent));
         let concrete = DiagnosticConcreteArgs::parse(ident, fields, attr, args)?;
-        Ok(DiagnosticDefArgs::Concrete(concrete))
+        Ok(DiagnosticDefArgs::Concrete(Box::new(concrete)))
     }
 }
 
@@ -177,73 +212,78 @@ impl Diagnostic {
             } => {
                 let (impl_generics, ty_generics, where_clause) = &generics.split_for_impl();
                 match args {
-                    DiagnosticDefArgs::Transparent => {
-                        if fields.iter().len() != 1 {
-                            return quote! {
-                                compile_error!("you can only use #[diagnostic(transparent)] on a struct with exactly one field");
-                            };
-                        }
-                        let field = fields
-                            .iter()
-                            .next()
-                            .expect("MIETTE BUG: thought we knew we had exactly one field");
-                        let field_name = field
-                            .ident
-                            .clone()
-                            .unwrap_or_else(|| format_ident!("unnamed"));
-                        let matcher = match fields {
-                            syn::Fields::Named(_) => quote! { let Self { #field_name } = self; },
-                            syn::Fields::Unnamed(_) => quote! { let Self(#field_name) = self; },
-                            syn::Fields::Unit => {
-                                unreachable!("MIETTE BUG: thought we knew we had exactly one field")
-                            }
-                        };
+                    DiagnosticDefArgs::Transparent(forward) => {
+                        let code_method = forward.gen_struct_method(WhichFn::Code);
+                        let help_method = forward.gen_struct_method(WhichFn::Help);
+                        let url_method = forward.gen_struct_method(WhichFn::Url);
+                        let labels_method = forward.gen_struct_method(WhichFn::Labels);
+                        let source_code_method = forward.gen_struct_method(WhichFn::SourceCode);
+                        let severity_method = forward.gen_struct_method(WhichFn::Severity);
+                        let related_method = forward.gen_struct_method(WhichFn::Related);
 
                         quote! {
                             impl #impl_generics miette::Diagnostic for #ident #ty_generics #where_clause {
-                                fn code<'a>(&'a self) -> std::option::Option<std::boxed::Box<dyn std::fmt::Display + 'a>> {
-                                    #matcher
-                                    #field_name.code()
-                                }
-                                fn help<'a>(&'a self) -> std::option::Option<std::boxed::Box<dyn std::fmt::Display + 'a>> {
-                                    #matcher
-                                    #field_name.help()
-                                }
-                                fn url<'a>(&'a self) -> std::option::Option<std::boxed::Box<dyn std::fmt::Display + 'a>> {
-                                    #matcher
-                                    #field_name.url()
-                                }
-                                fn severity(&self) -> std::option::Option<miette::Severity> {
-                                    #matcher
-                                    #field_name.severity()
-                                }
-                                fn snippets(&self) -> std::option::Option<std::boxed::Box<dyn std::iter::Iterator<Item = miette::DiagnosticSnippet> + '_>> {
-                                    #matcher
-                                    #field_name.snippets()
-                                }
+                                #code_method
+                                #help_method
+                                #url_method
+                                #labels_method
+                                #severity_method
+                                #source_code_method
+                                #related_method
                             }
                         }
                     }
                     DiagnosticDefArgs::Concrete(concrete) => {
-                        let code_body = concrete.code.as_ref().and_then(|x| x.gen_struct());
-                        let help_body = concrete.help.as_ref().and_then(|x| x.gen_struct(fields));
-                        let sev_body = concrete.severity.as_ref().and_then(|x| x.gen_struct());
-                        let snip_body = concrete
-                            .snippets
+                        let forward = |which| {
+                            concrete
+                                .forward
+                                .as_ref()
+                                .map(|fwd| fwd.gen_struct_method(which))
+                        };
+                        let code_body = concrete
+                            .code
                             .as_ref()
-                            .and_then(|x| x.gen_struct(fields));
+                            .and_then(|x| x.gen_struct())
+                            .or_else(|| forward(WhichFn::Code));
+                        let help_body = concrete
+                            .help
+                            .as_ref()
+                            .and_then(|x| x.gen_struct(fields))
+                            .or_else(|| forward(WhichFn::Help));
+                        let sev_body = concrete
+                            .severity
+                            .as_ref()
+                            .and_then(|x| x.gen_struct())
+                            .or_else(|| forward(WhichFn::Severity));
+                        let rel_body = concrete
+                            .related
+                            .as_ref()
+                            .and_then(|x| x.gen_struct())
+                            .or_else(|| forward(WhichFn::Related));
                         let url_body = concrete
                             .url
                             .as_ref()
-                            .and_then(|x| x.gen_struct(ident, fields));
-
+                            .and_then(|x| x.gen_struct(ident, fields))
+                            .or_else(|| forward(WhichFn::Url));
+                        let labels_body = concrete
+                            .labels
+                            .as_ref()
+                            .and_then(|x| x.gen_struct(fields))
+                            .or_else(|| forward(WhichFn::Labels));
+                        let src_body = concrete
+                            .source_code
+                            .as_ref()
+                            .and_then(|x| x.gen_struct(fields))
+                            .or_else(|| forward(WhichFn::SourceCode));
                         quote! {
                             impl #impl_generics miette::Diagnostic for #ident #ty_generics #where_clause {
                                 #code_body
                                 #help_body
                                 #sev_body
-                                #snip_body
+                                #rel_body
                                 #url_body
+                                #labels_body
+                                #src_body
                             }
                         }
                     }
@@ -258,14 +298,18 @@ impl Diagnostic {
                 let code_body = Code::gen_enum(variants);
                 let help_body = Help::gen_enum(variants);
                 let sev_body = Severity::gen_enum(variants);
-                let snip_body = Snippets::gen_enum(variants);
+                let labels_body = Labels::gen_enum(variants);
+                let src_body = SourceCode::gen_enum(variants);
+                let rel_body = Related::gen_enum(variants);
                 let url_body = Url::gen_enum(ident, variants);
                 quote! {
                     impl #impl_generics miette::Diagnostic for #ident #ty_generics #where_clause {
                         #code_body
                         #help_body
                         #sev_body
-                        #snip_body
+                        #labels_body
+                        #src_body
+                        #rel_body
                         #url_body
                     }
                 }
